// ------------------------------------------------------------
//
// [ptytunnel.c] tunnel on a terminal line
//
// Supposed to be a prove of concept.
//
// Copyright (c) 2006 by Lars Dlle <lars.doelle@on-line.de>
// All rights reserved.
//
// This program is free software; you can redistribute it
// and/or modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
// Boston, MA 02110-1301 USA
//
// ------------------------------------------------------------

#include <pty.h>
#include <utmp.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <stdarg.h>

int running = 1; // flag to stop the main loop
int master = -1; // master fd
pid_t child;     // child process pid

// ------------------------------------------------------------
//
// Logging
//
// ------------------------------------------------------------

FILE* log_file = 0;

static void logprintf(const char* fmt, ...)
{
  if (!log_file) return;
  va_list ap;
  va_start(ap, fmt);
  vfprintf(log_file, fmt, ap);
  va_end(ap);
  fflush(log_file);                 
}

#define LOG_HERE logprintf("%s(%d): here.\n",__FILE__,__LINE__)

// ------------------------------------------------------------
//
// Stream Processing
//
// Following the actual pipelines.
//
// ------------------------------------------------------------

int curr_dst   = 1; // can be fileno(stdout) or fileno(sysprt)
int printer_on = 0; // whether we are currently printing
FILE* sysprt   = 0; // printer pipe

void write_out(int dst, int cc)
{ int rc;

  if (dst < 0) return; // destination is dead.

Retry:
  rc = write(dst,&cc,1);
  if (rc == EINTR) goto Retry;

  if (rc <= 0)
  {
    // output failed, see which.
    if (printer_on)
    {
      // its the printer, further ignore, but keep
      // being in print mode
      logprintf("printer terminated\n");
      if (sysprt) pclose(sysprt);
      sysprt = 0;
      curr_dst = -1;
    }
    else
    {
      // its the terminal line, terminate.
      running = 0;
    }
  }
}

void write_pre(int fd, const char* prefix, int cc)
{
  while (*prefix) write_out(fd,*prefix++);
  write_out(fd,cc);
}

void do_printer(int dst, int on)
{
  if (on && printer_on)
  {
    // printer on while printing
    // issue "ESC[5i" to printer
    write_pre(dst,"\x1b[5",'i');
  }
  if (!on && !printer_on)
  {
    // printer off while not printing
    // issue "ESC[4i" to terminal
    write_pre(dst,"\x1b[4",'i');
  }
  if (on && !printer_on)
  {
    const char* command = getenv("PTYTUNNEL_PRINTER");
    if (command)
    {
      // start printer
      logprintf("starting printer '%s'\n",command);
      sysprt = popen(command,"w");
      curr_dst = fileno(sysprt);
    }
    else
    {
      logprintf("starting fake printer\n");
      curr_dst = -1;
    }
  }
  if (!on && printer_on)
  {
    // stop printer
    logprintf("stoping printer\n");
    if (sysprt) pclose(sysprt);
    sysprt = 0;
    curr_dst = fileno(stdout);
  }
  printer_on = on;
}

int state = 0; // state of the filter, implicit buffer.

void filter_upstream(int dst, int cc)
{
//logprintf("filter_upstream - state:%d, char:%c\n",state,cc);
  switch(state)
  {
    case 0:  // prefix: ""
             if (cc == 0x1b)               state =  1;
             else
             { write_pre(dst,"",cc);       state =  0; }
             break;
    case 1:  // prefix: "ESC"
             if (cc == '[' )               state =  2;
             else
             { write_pre(dst,"\x1b",cc);   state =  0; }
             break;
    case 2:  // prefix: "ESC["
             if (cc == '4' )               state = 34;
             else
             if (cc == '5' )               state = 35;
             else
             { write_pre(dst,"\x1b[",cc);  state =  0; }
             break;
    case 34: // prefix: "ESC[4"
             if (cc == 'i' )
             { do_printer(dst,0);          state =  0; }
             else
             { write_pre(dst,"\x1b[4",cc); state =  0; }
             break;
    case 35: // prefix: "ESC[5"
             if (cc == 'i' )
             { do_printer(dst,1);          state =  0; }
             else
             { write_pre(dst,"\x1b[5",cc); state =  0; }
             break;
  }
}

void filter_downstream(int dst, int cc)
{
  write_out(dst,cc); // identical stream
}

static void passdata(int fdsrc)
{
  char cc; int rc;

  rc = read(fdsrc,&cc,1);
  if (rc == EINTR) return;
  if (rc <= 0)
  {
    logprintf("read failed on %d - %s\n",fdsrc,strerror(errno));
    running = 0;
    return;
  }

  if (fdsrc == 0) // stdin
    filter_downstream(master,  cc);
  else
    filter_upstream  (curr_dst,cc);
}

// ------------------------------------------------------------
//
// PTY Pipeline
//
// ------------------------------------------------------------

static void window_size_change_handler(int signum)
{
  int serrno = errno;
  struct winsize wsize;
  ioctl(0,TIOCGWINSZ,(char *)&wsize);
  ioctl(master,TIOCSWINSZ,(char *)&wsize);
  errno = serrno;
}

static void child_termination_handler(int signum)
{
  int status;
  int serrno = errno;
  for(;;)
  {
    int pid = waitpid(WAIT_ANY, &status, WNOHANG);
    if (pid <= 0) break;
    logprintf(
      "received signal %d. Child %d terminating with status 0x%08x.\n",
      signum,pid,status
    );
  }
  errno = serrno;
}

void setSignalHandler()
{
  signal(SIGPIPE, SIG_IGN);
  signal(SIGWINCH, window_size_change_handler);
  signal(SIGCHLD, child_termination_handler);
}

struct termios termios_sav;

void ptytunnel(char* const argv[])
{
  int rc;

  tcgetattr(0, &termios_sav);

  struct winsize wsize;
  rc = ioctl(0,TIOCGWINSZ,(char *)&wsize);

  child = forkpty(&master,0,0,&wsize);
  if (child < 0) // error
  {
    fprintf(stderr,"fork failed - %s\n",strerror(errno));
    exit(1);
  }
  if (child == 0) // child process
  {
    execvp(argv[0],argv);
    // we failed, terminate.
    fprintf(stderr,"exec failed for program '%s' - %s\n",argv[0],strerror(errno));
    exit(1);
  }
}

#define MAX(A,B) ((A)>(B)?(A):(B))

static void runtunnel()
{
  setSignalHandler();

  struct termios termios_now;
  tcgetattr(0, &termios_now);
  cfmakeraw(&termios_now);
  tcsetattr(0, TCSANOW, &termios_now);

  while (running)
  {
    // setup select parameters

    fd_set rd; FD_ZERO(&rd);

    int maxfd = -1;

    FD_SET(master,&rd); maxfd = MAX(master,maxfd);
    FD_SET(0     ,&rd); maxfd = MAX(0     ,maxfd);

    int rc = select(maxfd+1,&rd,0,0,0);

    if (rc < 0 && errno != EINTR) { fprintf(stderr,"select %s\n",strerror(errno)); break; }
    if (rc < 0 && errno == EINTR) continue;

    if (FD_ISSET(master,&rd)) passdata(master);
    if (FD_ISSET(0,     &rd)) passdata(0);
  }

  tcsetattr(0, TCSANOW, &termios_sav);
}

// ------------------------------------------------------------
//
// MAIN
//
// ------------------------------------------------------------

#define COPYRIGHT \
"ptytunnel - tunnel on a terminal line (Version " VERSION ")\n" \
"\n" \
"Copyright (c) 2006 by Lars Doelle <lars.doelle@on-line.de>\n" \
"All rights reserved.\n" \
"\n" \
"This program is free software; you can redistribute it\n" \
"and/or modify it under the terms of the GNU General Public\n" \
"License as published by the Free Software Foundation; either\n" \
"version 2 of the License, or (at your option) any later\n" \
"version.\n" \
"\n" \
"This program is distributed in the hope that it will be\n" \
"useful, but WITHOUT ANY WARRANTY; without even the implied\n" \
"warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR\n" \
"PURPOSE.  See the GNU General Public License for more\n" \
"details.\n" \
"\n" \
"You should have received a copy of the GNU General Public\n" \
"License along with this program; if not, write to the Free\n" \
"Software Foundation, Inc., 51 Franklin Street, Fifth Floor,\n" \
"Boston, MA 02110-1301 USA\n" 

void usage()
{
  fprintf(stderr,"usage: ptytunnel [options] command arg ...\n");
}

int main(int argc, char* const argv[])
{
  if (!isatty(0) || !isatty(1))
  {
    fprintf(stderr,"error - not running on a tty\n");
    return 1;
  }

  int i;
  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i],"-v"))
    {
      fprintf(stderr,"%s",COPYRIGHT);
      return 0;
    }

    if (!strcmp(argv[i],"-h")) { usage(); return 0; }
    if (argv[i][0] == '-')
    {
      fprintf(stderr,"error: unknown option '%s'.\n",argv[i]);
      usage();
      return 1;
    }
    else break;
  }

  if (i == argc)
  {
    char* const args[] = { "/bin/sh", 0 };
    ptytunnel(args);
  }
  else
  {
    ptytunnel(argv+i);
  }

//log_file = fopen("ptytunnel.log","w");

  runtunnel();

  if (log_file) fclose(log_file);

  return 0;
}
